#include <assert.h>

// Merge a[begin..mid - 1] and a[mid..end] into b[begin..end]
void merge(int a[], int b[], int begin, int mid, int end) {
    int i_begin = begin;
    int i_mid   = mid;
    for (int j = begin; j < end; j++)
        if (i_begin < mid && (i_mid >= end || a[i_begin] <= a[i_mid]))
            b[j] = a[i_begin++];
        else
            b[j] = a[i_mid++];
}

// Copy b[begin..end] into a[begin..end]
void copy(int a[], int b[], int begin, int end) {
    for (int k = begin; k < end; k++)
        a[k] = b[k];
}

// Sort a[begin..end], using b[begin..end] as buffer
void partial_merge_sort(int a[], int b[], int begin, int end)
{
    if (end - begin < 2)
        return;
    
    printf("Sorting [");
    for (int j = begin; j < end; j++)
        printf("%d ", a[j]);
    printf("]\n");

    // Split and sort
    int mid = begin + (end - begin) / 2;
    partial_merge_sort(a, b, begin, mid);
    partial_merge_sort(a, b, mid, end);
    
    // Merge and copy
    merge(a, b, begin, mid, end);
    copy(a, b, begin, end);
}

// Sort a[0..size], using b[0..size] as a buffer
void merge_sort(int a[], int b[], int size)
{
    partial_merge_sort(a, b, 0, size);
}

int unsorted_a[] = {10, -10, 7, 2, 2, -4, -7, -10, 1, 4, 5};
int buffer[] = {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0};

int sorted_a[]   = {-10, -10, -7, -4, 1, 2, 2, 4, 5, 7, 10};

void test_merge_sort() {
    int size = sizeof(unsorted_a) / sizeof(unsorted_a[0]);

    merge_sort(unsorted_a, buffer, size);
    for (int i = 0; i < size; i++)
        assert(sorted_a[i] == unsorted_a[i]);
}

int main(int argc, char *argv[])
{
    test_merge_sort();
}